import { __values, __spread } from 'tslib';
import { keys, isUndefined, last, toString } from 'micro-dash';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Extend this for classes whose objects are directly callable. Sadly, you'll need to define an extra interface and repeat the typing information, for now. (Maybe fixable with TypeScript 3 tuples?)
 *
 * ```ts
 * interface Multiplier {
 *   (value: number): number;
 * }
 *
 * class Multiplier extends CallableObject {
 *   constructor(public factor: number) {
 *     super((value: number) => value * this.factor);
 *   }
 * }
 *
 * const doubler = new Multiplier(2);
 * doubler(2); // 4
 * doubler.factor = 3;
 * doubler(2); // 6
 * ```
 * @abstract
 */
var  /**
 * Extend this for classes whose objects are directly callable. Sadly, you'll need to define an extra interface and repeat the typing information, for now. (Maybe fixable with TypeScript 3 tuples?)
 *
 * ```ts
 * interface Multiplier {
 *   (value: number): number;
 * }
 *
 * class Multiplier extends CallableObject {
 *   constructor(public factor: number) {
 *     super((value: number) => value * this.factor);
 *   }
 * }
 *
 * const doubler = new Multiplier(2);
 * doubler(2); // 4
 * doubler.factor = 3;
 * doubler(2); // 6
 * ```
 * @abstract
 */
CallableObject = /** @class */ (function () {
    function CallableObject(f) {
        var _newTarget = this.constructor;
        return Object.setPrototypeOf(f, _newTarget.prototype);
    }
    return CallableObject;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
var seq = 0;
/**
 * Creates a factory function for building objects of a given type. Commonly used to create test objects for use in specs.
 *
 * ```ts
 * interface Message {
 *   id: number;
 *   text: string;
 * }
 *
 * interface Options {
 *   cypherDistance: number;
 * }
 *
 * declare function shiftCharacters(text: string, distance: number): string;
 *
 * const buildMessage = createBuilder<Message, Options>(
 *   (seq) => ({ id: seq, text: `message ${seq}` }),
 *   (message, _seq, options) => {
 *     message.text = shiftCharacters(message.text, options.cypherDistance || 0);
 *   },
 * );
 *
 * buildMessage(); // { id: 1, text: "message 1" }
 * buildMessage({ text: "Hello world!" }); // { id: 2, text: "Hello world!" }
 * buildMessage({ text: "abc" }, { cypherDistance: 3 }); // { id: 3, text: "def" }
 * ```
 * @template T, OptionsType
 * @param {?} buildDefaults
 * @param {?=} afterBuild
 * @return {?}
 */
function createBuilder(buildDefaults, afterBuild) {
    return (/**
     * @param {?=} attributes
     * @param {?=} options
     * @return {?}
     */
    function (attributes, options) {
        if (options === void 0) { options = {}; }
        ++seq;
        /** @type {?} */
        var obj = Object.assign(buildDefaults(seq, options), attributes);
        if (afterBuild) {
            afterBuild(obj, seq, options);
        }
        return obj;
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * An object you can use to create a promise and resolve/reject it later.
 *
 * ```ts
 * const deferred = new Deferred<string>();
 *
 * // later
 * deferred.resolve('a nice value'); // causes `deferred.promise` to resolve
 *
 * // or
 * deferred.reject('a bad value'); // causes `deferred.promise` to reject
 * ```
 * @template T
 */
var  /**
 * An object you can use to create a promise and resolve/reject it later.
 *
 * ```ts
 * const deferred = new Deferred<string>();
 *
 * // later
 * deferred.resolve('a nice value'); // causes `deferred.promise` to resolve
 *
 * // or
 * deferred.reject('a bad value'); // causes `deferred.promise` to reject
 * ```
 * @template T
 */
Deferred = /** @class */ (function () {
    function Deferred() {
        var _this = this;
        this.promise = new Promise((/**
         * @param {?} resolve
         * @param {?} reject
         * @return {?}
         */
        function (resolve, reject) {
            _this.resolve = resolve;
            _this.reject = reject;
        }));
    }
    return Deferred;
}());

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Performs a deep comparison between two values to determine if they are equivalent, up to the given depth. Once that depth is reached, remaining values are compared using `Object.is()`.
 *
 * ```ts
 * let object1 = "a";
 * let object2 = "a";
 * isEqualAtDepth(0, object1, object2); // true
 * isEqualAtDepth(1, object1, object2); // true
 *
 * let object1 = { a: 1 };
 * let object2 = { a: 1 };
 * isEqualAtDepth(0, object1, object2); // false
 * isEqualAtDepth(1, object1, object2); // true
 *
 * object1 = { a: [1, 2, 3], d: { e: 1 } };
 * object2 = { a: [1, 2, 3], d: { e: 1 } };
 * isEqualAtDepth(0, object1, object2); // false
 * isEqualAtDepth(1, object1, object2); // false
 * isEqualAtDepth(2, object1, object2); // true
 * isEqualAtDepth(3, object1, object2); // true
 * ```
 * @param {?} depth
 * @param {?} value
 * @param {?} other
 * @return {?}
 */
function isEqualAtDepth(depth, value, other) {
    var e_1, _a, e_2, _b;
    if (Object.is(value, other)) {
        // covers e.g. NaN === NaN
        return true;
    }
    if (depth === 0 || !(value instanceof Object && other instanceof Object)) {
        return false;
    }
    try {
        for (var _c = __values(keys(value)), _d = _c.next(); !_d.done; _d = _c.next()) {
            var key = _d.value;
            if (!other.hasOwnProperty(key)) {
                return false;
            }
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (_d && !_d.done && (_a = _c.return)) _a.call(_c);
        }
        finally { if (e_1) throw e_1.error; }
    }
    try {
        for (var _e = __values(keys(other)), _f = _e.next(); !_f.done; _f = _e.next()) {
            var key = _f.value;
            if (!isEqualAtDepth(depth - 1, value[key], other[key])) {
                return false;
            }
        }
    }
    catch (e_2_1) { e_2 = { error: e_2_1 }; }
    finally {
        try {
            if (_f && !_f.done && (_b = _e.return)) _b.call(_e);
        }
        finally { if (e_2) throw e_2.error; }
    }
    return true;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Checks if `value` is anything other than `undefined`.
 *
 * ```ts
 * isDefined({ a: 1 }); // true
 * isDefined(false); // true
 * isDefined(null); // true
 * isDefined(undefined); // false
 * ```
 * @param {?} value
 * @return {?}
 */
function isDefined(value) {
    return !isUndefined(value);
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Rounds `value` to the nearest multiple of `multiple`.
 *
 * ```ts
 * roundToMultiple(5, 13); // 15
 * roundToMultiple(2, 4.8); // 4
 * roundToMultiple(3, -4); // -3
 * roundToMultiple(0.5, 1.6); // 1.5
 * ```
 * @param {?} multiple
 * @param {?} value
 * @return {?}
 */
function roundToMultipleOf(multiple, value) {
    return Math.round(value / multiple) * multiple;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Returns a promise that resolves after `ms` milliseconds.
 *
 * ```ts
 * // do something
 * await sleep(1000); // wait a second
 * // do something else
 * ```
 * @param {?} ms
 * @return {?}
 */
function sleep(ms) {
    return new Promise((/**
     * @param {?} resolve
     * @return {?}
     */
    function (resolve) {
        setTimeout(resolve, ms);
    }));
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
var nanoConversions = {};
/** @enum {string} */
var TimeUnit = {
    Nanoseconds: "ns",
    Microseconds: "μs",
    Milliseconds: "ms",
    Seconds: "s",
    Minutes: "m",
    Hours: "h",
    Days: "d",
    Weeks: "w",
    Years: "y",
    Decades: "dec",
    Centuries: "cent",
    Millennia: "mil",
};
/**
 * Converts time between two units. Units can be any value described in the docs for [[TimeUnit]].
 *
 * ```ts
 * convertTime(1, 's', 'ms'); // 1000
 * convertTime(20, TimeUnit.Decades, TimeUnit.Centuries); // 2
 * ```
 * @param {?} value
 * @param {?} unit
 * @param {?} targetUnit
 * @return {?}
 */
function convertTime(value, unit, targetUnit) {
    return (value * nanoConversions[unit]) / nanoConversions[targetUnit];
}
/**
 * Constructs a string representation of an elapsed amount of time. The least significant unit will be rounded to the nearest whole number.
 *
 * ```ts
 * elapsedToString(2001, ["s", "ms"]); // "2 s 1 ms"
 * elapsedToString(15, ["wks", "d"], ( elapsedUnit: TimeUnit.Days }); // "2 wks 1 d"
 * elapsedToString(1, [TimeUnit.Microseconds]); // "1000 μs"
 * ```
 *
 * @param {?} elapsed
 * @param {?} units
 * @param {?=} __2
 * @return {?}
 */
function elapsedToString(elapsed, units, _a) {
    var _b = _a === void 0 ? {} : _a, _c = _b.elapsedUnit, elapsedUnit = _c === void 0 ? TimeUnit.Milliseconds : _c, _d = _b.showLeadingZeros, showLeadingZeros = _d === void 0 ? true : _d;
    elapsed = roundToMultipleOf(convertTime(1, last(units), elapsedUnit), elapsed);
    /** @type {?} */
    var showZeros = showLeadingZeros;
    /** @type {?} */
    var tokens = [];
    for (var i = 0; i < units.length; ++i) {
        /** @type {?} */
        var unit = units[i];
        /** @type {?} */
        var conversion = convertTime(1, elapsedUnit, unit);
        /** @type {?} */
        var value = Math.floor(elapsed * conversion);
        if (value > 0 || showZeros || i === units.length - 1) {
            tokens.push(value);
            tokens.push(unit);
            showZeros = true;
        }
        elapsed -= value / conversion;
    }
    return tokens.join(" ");
}
addConversion(TimeUnit.Nanoseconds, "Nanosecond", 1, { aliases: ["nanos"] });
addConversion(TimeUnit.Microseconds, "Microsecond", 1000, {
    aliases: ["micros"],
});
addConversion(TimeUnit.Milliseconds, "Millisecond", 1000 * 1000, {
    aliases: ["millis"],
});
addConversion(TimeUnit.Seconds, "Second", 1000 * 1000 * 1000, {
    aliases: ["S", "sec", "secs"],
});
addConversion(TimeUnit.Minutes, "Minute", 60 * 1000 * 1000 * 1000, {
    aliases: ["M", "min", "mins"],
});
addConversion(TimeUnit.Hours, "Hour", 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["H", "hr", "hrs"],
});
addConversion(TimeUnit.Days, "Day", 24 * 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["D"],
});
addConversion(TimeUnit.Weeks, "Week", 7 * 24 * 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["W", "wk", "wks"],
});
addConversion(TimeUnit.Years, "Year", 365 * 24 * 60 * 60 * 1000 * 1000 * 1000, {
    aliases: ["Y", "yr", "yrs"],
});
addConversion(TimeUnit.Decades, "Decade", 10 * 365 * 24 * 60 * 60 * 1000 * 1000 * 1000);
addConversion(TimeUnit.Centuries, "Century", 100 * 365 * 24 * 60 * 60 * 1000 * 1000 * 1000, { plural: "Centuries" });
addConversion(TimeUnit.Millennia, "Millennium", 1000 * 365 * 24 * 60 * 60 * 1000 * 1000 * 1000, { plural: "Millennia" });
/**
 * @param {?} unit
 * @param {?} singular
 * @param {?} nanos
 * @param {?=} __3
 * @return {?}
 */
function addConversion(unit, singular, nanos, _a) {
    var e_1, _b;
    var _c = _a === void 0 ? {} : _a, _d = _c.aliases, aliases = _d === void 0 ? (/** @type {?} */ ([])) : _d, _e = _c.plural, plural = _e === void 0 ? singular + "s" : _e;
    try {
        for (var _f = __values(__spread([
            unit,
            singular,
            singular.toLowerCase(),
            plural,
            plural.toLowerCase()
        ], aliases)), _g = _f.next(); !_g.done; _g = _f.next()) {
            var key = _g.value;
            nanoConversions[key] = nanos;
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (_g && !_g.done && (_b = _f.return)) _b.call(_f);
        }
        finally { if (e_1) throw e_1.error; }
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Converts a 2D array to a csv string. Values are converted using micro-dash's toString().
 *
 * ```ts
 * toCsv([["a", "b", "c"], ["d", "e", "f"], ["g", "h", "i"]]); // "a,b,c\nd,e,f\ng,h,i"
 * toCsv([
 *   ["a", "", "string"]
 *   [undefined, null],
 *   [true, false],
 *   [1, 2, 3],
 *   [{}, { hi: "there" }]
 * ]) // "a,,string\n,\ntrue,false\n1,2,3\n[object Object],[object Object]"
 * ```
 * @param {?} content
 * @return {?}
 */
function toCsv(content) {
    return content
        .map((/**
     * @param {?} row
     * @return {?}
     */
    function (row) { return row.map((/**
     * @param {?} cell
     * @return {?}
     */
    function (cell) { return toCellString(cell); })).join(","); }))
        .join("\n");
}
/** @type {?} */
var specialCsvCharactersRegexp = /["|,|\n|\r]/;
/** @type {?} */
var allDoubleQuotes = /"/g;
/**
 * @param {?} value
 * @return {?}
 */
function toCellString(value) {
    /** @type {?} */
    var str = toString(value);
    if (specialCsvCharactersRegexp.test(str)) {
        return "\"" + str.replace(allDoubleQuotes, "\"\"") + "\"";
    }
    else {
        return str;
    }
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * Returns a new function to use in place of `func` that will call the provided hooks in addition to `func`. They are called in the following order:
 *
 * 1. `before`
 * 2. `original`
 * 3. `transform`
 * 4. `after`
 *
 * ```ts
 * const sum = (a: number, b: number) => a + b;
 * const sumAndLog = wrapFunction(sum, {
 *   after: (result, a, b) => {
 *     console.log(a, '+', b, '=', result);
 *   },
 * }
 * const sumPlusOne = wrapFunction(sum, {
 *   transform: (result) => result + 1,
 * });
 * ```
 * @template A, R, T
 * @param {?} original
 * @param {?} __1
 * @return {?}
 */
function wrapFunction(original, _a) {
    var before = _a.before, transform = _a.transform, after = _a.after;
    /** @type {?} */
    var wrapped = (/**
     * @this {?}
     * @param {...?} args
     * @return {?}
     */
    function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        if (before) {
            before.apply(this, args);
        }
        /** @type {?} */
        var result = original.apply(this, args);
        if (transform) {
            result = ((/** @type {?} */ (transform))).call.apply(((/** @type {?} */ (transform))), __spread([this, result], args));
        }
        if (after) {
            ((/** @type {?} */ (after))).call.apply(((/** @type {?} */ (after))), __spread([this, result], args));
        }
        return result;
    });
    Object.defineProperty(wrapped, "length", { value: original.length });
    return wrapped;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingOverride,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { CallableObject, Deferred, TimeUnit, convertTime, createBuilder, elapsedToString, isDefined, isEqualAtDepth, roundToMultipleOf, sleep, toCsv, wrapFunction };
//# sourceMappingURL=s-js-utils.js.map
